@@ -26,7 +26,7 @@
  */
  
 #include "thdataleg.h"
-
+#include "thcs.h"
 
 thdataleg::thdataleg()
 {
@@ -46,6 +46,7 @@ void thdataleg::clear()
   this->psurvey = NULL;
   this->s_mark = TT_DATAMARK_UNKNOWN;
   this->flags = TT_LEGFLAG_NONE;
+  this->gridcs = TTCS_LOCAL;
   
   this->station.clear();
   this->from.clear();
@@ -94,6 +95,7 @@ void thdataleg::clear()
   this->y_sd = thnan;
   this->z_sd = thnan;
   this->declination = thnan;
+  this->implicit_declination = 0.0;
   
   this->from_up = thnan;
   this->from_down = thnan;
@@ -106,8 +108,10 @@ void thdataleg::clear()
   this->to_right = thnan;
   
   this->walls = TT_AUTO;
+  this->splay_walls = true;
   this->shape = TT_DATALEG_SHAPE_UNKNOWN;
   this->vtresh = thnan;
+  this->extend_ratio = thnan;
     
   this->infer_plumbs = false;
   this->infer_equates = false;
@@ -116,6 +120,9 @@ void thdataleg::clear()
   this->to_be_adjusted = false;
   this->topofil = false;
   this->plumbed = false;
+
+  this->loop = NULL;
+  this->traverse =  NULL;
 }
 
 
@@ -132,6 +139,7 @@ thdatafix::thdatafix()
   this->cxy = thnan;
   this->cyz = thnan;
   this->cxz = thnan;
+  this->cs = TTCS_LOCAL;
 }
 
 
@@ -141,6 +149,7 @@ thdatass::thdatass()
   this->comment = NULL;
   this->flags = TT_STATIONFLAG_NONE;
   this->psurvey = NULL;
+  this->explored = thnan;
 }
 
 
@@ -165,13 +174,12 @@ void thdataleg::calc_total_stds()
 //    }
 //  }
   
-  double dx, dy, dz, dL, dl, dT, dC, ddx, ddy, ddz, dh, dZ;
+  double dx, dy, dz, dL, dl, dT, dC, ddx, ddy, ddz, dZ;
 #define setd(a,b,v) if (thisnan(b)) a = v; else a = b;
   setd(dx, this->x_sd, 0.057735026919);
   setd(dy, this->y_sd, 0.057735026919);
   setd(dz, this->z_sd, 0.057735026919);
   setd(dL, this->length_sd, 0.1);
-  setd(dh, this->length_sd, 0.1);
   setd(dl, this->counter_sd, 0.1);
   setd(dZ, this->depth_sd, 0.1);
   setd(dT, this->bearing_sd, 1.0);
@@ -193,7 +201,7 @@ void thdataleg::calc_total_stds()
   T = this->total_bearing / 180.0 * 3.14159265359;
   C = this->total_gradient / 180.0 * 3.14159265359;
   L = this->total_length;
-  h = hypot(this->total_dx, this->total_dy);
+  h = std::hypot(this->total_dx, this->total_dy);
 
   if (this->topofil) {
     dL = 1.41421356237 * dl;
@@ -253,3 +261,11 @@ void thdataleg::calc_total_stds()
   }
 }
 
+void thdataleg::export_mp_flags(FILE * out)
+{
+	fprintf(out, "ATTR__shotflag_splay := %s;\n", (((this->flags & TT_LEGFLAG_SPLAY) != 0) ? "true" : "false"));
+	fprintf(out, "ATTR__shotflag_duplicate := %s;\n", (((this->flags & TT_LEGFLAG_DUPLICATE) != 0) ? "true" : "false"));
+	fprintf(out, "ATTR__shotflag_approx := %s;\n", (((this->flags & TT_LEGFLAG_APPROXIMATE) != 0) ? "true" : "false"));
+}
+
+
