@@ -32,7 +32,10 @@
 #include "thparse.h"
 #include "thobjectname.h"
 #include "thobjectsrc.h"
+#include "thinfnan.h"
 #include <list>
+#include <string>
+#include <map>
 
 /**
  * survey leg components.
@@ -72,7 +75,9 @@ enum {
   TT_DATALEG_DOWN,
   TT_DATALEG_LEFT,
   TT_DATALEG_RIGHT,
+  TT_DATALEG_DIMS,
   TT_DATALEG_IGNORE,
+  TT_DATALEG_IGNOREALL,
 };
 
 
@@ -83,7 +88,7 @@ enum {
  
 static const thstok thtt_dataleg_comp[] = {
   {"altitude", TT_DATALEG_ALTITUDE},
-  {"assistent", TT_DATALEG_ASSISTANT},
+  {"assistant", TT_DATALEG_ASSISTANT},
   {"backbearing", TT_DATALEG_BACKBEARING},
   {"backclino", TT_DATALEG_BACKGRADIENT},
   {"backcompass", TT_DATALEG_BACKBEARING},
@@ -96,6 +101,7 @@ static const thstok thtt_dataleg_comp[] = {
   {"counter", TT_DATALEG_COUNT},
   {"depth", TT_DATALEG_DEPTH},
   {"depthchange", TT_DATALEG_DEPTHCHANGE},
+  {"dimensions", TT_DATALEG_DIMS},
   {"direction", TT_DATALEG_DIRECTION},
   {"dog", TT_DATALEG_ASSISTANT},
   {"down", TT_DATALEG_DOWN},
@@ -110,6 +116,7 @@ static const thstok thtt_dataleg_comp[] = {
   {"fromdepth", TT_DATALEG_FROMDEPTH},
   {"gradient", TT_DATALEG_GRADIENT},
   {"ignore", TT_DATALEG_IGNORE},
+  {"ignoreall", TT_DATALEG_IGNOREALL},
   {"instruments", TT_DATALEG_INSTRUMENTS},
   {"insts", TT_DATALEG_INSTRUMENTS},
   {"left", TT_DATALEG_LEFT},
@@ -118,6 +125,7 @@ static const thstok thtt_dataleg_comp[] = {
   {"northing", TT_DATALEG_NORTHING},
   {"notebook", TT_DATALEG_NOTES},
   {"notes", TT_DATALEG_NOTES},
+  {"pics", TT_DATALEG_PICTURES},
   {"pictures", TT_DATALEG_PICTURES},
   {"position", TT_DATALEG_POSITION},
   {"right", TT_DATALEG_RIGHT},
@@ -147,7 +155,7 @@ enum {
 
 
 /**
- * Dataleg infer componets.
+ * Dataleg infer components.
  */
  
 static const thstok thtt_dataleg_infer[] = {
@@ -157,6 +165,30 @@ static const thstok thtt_dataleg_infer[] = {
 };
 
 
+/**
+ * Shape components.
+ */
+ 
+enum {
+  TT_DATALEG_SHAPE_DIAMOND,
+  TT_DATALEG_SHAPE_OCTAGON,
+  TT_DATALEG_SHAPE_RECTANGLE,
+  TT_DATALEG_SHAPE_TRIANGLE,
+  TT_DATALEG_SHAPE_TUNNEL,
+  TT_DATALEG_SHAPE_UNKNOWN,
+};
+
+static const thstok thtt_dataleg_shape[] = {
+  {"diamond", TT_DATALEG_SHAPE_DIAMOND},
+  {"octagon", TT_DATALEG_SHAPE_OCTAGON},
+  {"rect", TT_DATALEG_SHAPE_RECTANGLE},
+  {"rectangle", TT_DATALEG_SHAPE_RECTANGLE},
+  {"triangle", TT_DATALEG_SHAPE_TRIANGLE},
+  {"tunnel", TT_DATALEG_SHAPE_TUNNEL},
+  {NULL, TT_DATALEG_SHAPE_UNKNOWN},
+};
+
+
 /**
  * Data types.
  */
@@ -167,6 +199,7 @@ enum {
   TT_DATATYPE_CARTESIAN,
   TT_DATATYPE_CYLPOLAR,
   TT_DATATYPE_NOSURVEY,
+  TT_DATATYPE_DIMS,
   TT_DATATYPE_UNKNOWN,
 };
 
@@ -178,6 +211,7 @@ enum {
 static const thstok thtt_datatype[] = {
   {"cartesian", TT_DATATYPE_CARTESIAN},
   {"cylpolar", TT_DATATYPE_CYLPOLAR},
+  {"dimensions", TT_DATATYPE_DIMS},
   {"diving", TT_DATATYPE_DIVING},
   {"normal", TT_DATATYPE_NORMAL},
   {"nosurvey", TT_DATATYPE_NOSURVEY},
@@ -221,9 +255,42 @@ enum {
   TT_DATALFLAG_UNKNOWN,
   TT_DATALFLAG_SURFACE,
   TT_DATALFLAG_DUPLICATE,
+  TT_DATALFLAG_APPROXIMATE,
+  TT_DATALFLAG_SPLAY,
   TT_DATALFLAG_NOT,  
 };
 
+
+enum {
+  TT_EXTENDFLAG_UNKNOWN = 0,
+  TT_EXTENDFLAG_NORMAL = 1,
+  TT_EXTENDFLAG_REVERSE = 2,
+  TT_EXTENDFLAG_LEFT = 4,
+  TT_EXTENDFLAG_RIGHT = 8,
+  TT_EXTENDFLAG_VERTICAL = 16,
+  TT_EXTENDFLAG_DIRECTION = 31,
+  TT_EXTENDFLAG_BREAK = 32,
+  TT_EXTENDFLAG_START = 64,
+  TT_EXTENDFLAG_IGNORE = 128,
+  TT_EXTENDFLAG_HIDE = 256,
+  TT_EXTENDFLAG_CNDIGNORE = 512,
+};
+
+
+static const thstok thtt_extendflag[] = {
+  //{"break", TT_EXTENDFLAG_BREAK},
+  {"hide", TT_EXTENDFLAG_HIDE},
+  {"ignore", TT_EXTENDFLAG_IGNORE},
+  {"left", TT_EXTENDFLAG_LEFT},
+  {"normal", TT_EXTENDFLAG_NORMAL},
+  {"reverse", TT_EXTENDFLAG_REVERSE},
+  {"right", TT_EXTENDFLAG_RIGHT},
+  {"start", TT_EXTENDFLAG_START},
+  {"vertical", TT_EXTENDFLAG_VERTICAL},
+  {NULL, TT_EXTENDFLAG_UNKNOWN},
+};
+
+
 /**
  * Data leg flags.
  */
@@ -232,6 +299,8 @@ enum {
   TT_LEGFLAG_NONE = 0,
   TT_LEGFLAG_SURFACE = 1,
   TT_LEGFLAG_DUPLICATE = 2,
+  TT_LEGFLAG_APPROXIMATE = 4,
+  TT_LEGFLAG_SPLAY = 8,
 };
 
 
@@ -240,8 +309,11 @@ enum {
  */
  
 static const thstok thtt_datalflag[] = {
+  {"approx", TT_DATALFLAG_APPROXIMATE},
+  {"approximate", TT_DATALFLAG_APPROXIMATE},
   {"duplicate", TT_DATALFLAG_DUPLICATE},
   {"not", TT_DATALFLAG_NOT},
+  {"splay", TT_DATALFLAG_SPLAY},
   {"surface", TT_DATALFLAG_SURFACE},
   {NULL, TT_DATALFLAG_UNKNOWN},
 };
@@ -256,6 +328,19 @@ enum {
   TT_DATASFLAG_ENTRANCE,
   TT_DATASFLAG_CONT,
   TT_DATASFLAG_FIXED,
+  TT_DATASFLAG_NOTFIXED,
+  TT_DATASFLAG_ATTR,
+  TT_DATASFLAG_NOT,
+  TT_DATASFLAG_EXPLORED,
+  TT_DATASFLAG_AIRDRAUGHT,
+  TT_DATASFLAG_AIRDRAUGHT_WINTER,
+  TT_DATASFLAG_AIRDRAUGHT_SUMMER,
+  TT_DATASFLAG_SINK,
+  TT_DATASFLAG_SPRING,
+  TT_DATASFLAG_DOLINE,
+  TT_DATASFLAG_DIG,
+  TT_DATASFLAG_OVERHANG,
+  TT_DATASFLAG_ARCH,
 };
 
 /**
@@ -267,6 +352,17 @@ enum {
   TT_STATIONFLAG_ENTRANCE = 1,
   TT_STATIONFLAG_CONT = 2,
   TT_STATIONFLAG_FIXED = 4,
+  TT_STATIONFLAG_UNDERGROUND = 8,
+  TT_STATIONFLAG_NOTFIXED = 16,
+  TT_STATIONFLAG_SPRING = 32,
+  TT_STATIONFLAG_SINK = 64,
+  TT_STATIONFLAG_DOLINE = 128,
+  TT_STATIONFLAG_DIG = 256,
+  TT_STATIONFLAG_AIRDRAUGHT = 1536,
+  TT_STATIONFLAG_AIRDRAUGHT_WINTER = 512,
+  TT_STATIONFLAG_AIRDRAUGHT_SUMMER = 1024,
+  TT_STATIONFLAG_OVERHANG = 2048,
+  TT_STATIONFLAG_ARCH = 4096,
 };
 
 
@@ -275,8 +371,21 @@ enum {
  */
  
 static const thstok thtt_datasflag[] = {
+  {"air-draught", TT_DATASFLAG_AIRDRAUGHT},
+  {"air-draught:summer", TT_DATASFLAG_AIRDRAUGHT_SUMMER},
+  {"air-draught:winter", TT_DATASFLAG_AIRDRAUGHT_WINTER},
+  {"arch", TT_DATASFLAG_ARCH},
+  {"attr", TT_DATASFLAG_ATTR},
   {"continuation", TT_DATASFLAG_CONT},
+  {"dig", TT_DATASFLAG_DIG},
+  {"doline", TT_DATASFLAG_DOLINE},
   {"entrance", TT_DATASFLAG_ENTRANCE},
+  {"explored", TT_DATASFLAG_EXPLORED},
+  {"fixed", TT_DATASFLAG_FIXED},
+  {"not", TT_DATASFLAG_NOT},
+  {"overhang", TT_DATASFLAG_OVERHANG},
+  {"sink", TT_DATASFLAG_SINK},
+  {"spring", TT_DATASFLAG_SPRING},
   {NULL, TT_DATASFLAG_UNKNOWN},
 };
 
@@ -296,21 +405,31 @@ class thdataleg {
   int data_type,  ///< leg data type
       flags;  ///< Leg flags.
       
-  unsigned char s_mark;  ///< Type of the station mark
+  unsigned int s_mark,  ///< Type of the station mark
+    extend;  ///< Extend flags: normal, reverse, left, right, break
+
+  int walls, shape, gridcs;
   
+  bool splay_walls;
+
+  struct thdb1d_loop * loop; ///< Worst loop leg is a part of.
+  struct thdb1d_traverse * traverse; ///< Centreline traverse, leg is a part of.
+
   thobjectname station, from, to;
   class thsurvey * psurvey;  ///< parent survey
   
   double length, counter, fromcounter, tocounter, depth, fromdepth,
     todepth, depthchange, bearing, gradient, dx, dy, dz,
     backbearing, backgradient, total_length, total_bearing, total_gradient,
-    total_dx, total_dy, total_dz, adj_dx, adj_dy, adj_dz;
+    total_dx, total_dy, total_dz, adj_dx, adj_dy, adj_dz,
+    from_up, from_down, from_left, from_right,
+    to_up, to_down, to_left, to_right, vtresh, extend_ratio;
     
   double length_sd, counter_sd, depth_sd, bearing_sd, gradient_sd,
-    dx_sd, dy_sd, dz_sd, x_sd, y_sd, z_sd, declination, 
-    total_sdx, total_sdy, total_sdz;
+    dx_sd, dy_sd, dz_sd, x_sd, y_sd, z_sd, declination, implicit_declination, 
+    total_sdx, total_sdy, total_sdz, fxx, txx;
     
-  bool infer_plumbs, infer_equates, direction, adjusted, topofil;
+  bool infer_plumbs, infer_equates, direction, adjusted, to_be_adjusted, topofil, plumbed;
   
  
   /**
@@ -334,6 +453,14 @@ class thdataleg {
   void clear();
   
   void calc_total_stds();
+
+
+  /**
+   * Export station attributes to metapost.
+   */
+
+  void export_mp_flags(FILE * out);
+
   
 };
 
@@ -345,6 +472,8 @@ class thdataleg {
 class thdatafix {
   
   public:
+
+  int cs; ///< Fix CS.
   
   thobjectname station;  ///< Station name.
 
@@ -359,6 +488,9 @@ class thdatafix {
 };
 
 
+typedef std::map<std::string, const char *> thdatass_attr_map;
+
+
 /**
  * Station class.
  */
@@ -373,9 +505,13 @@ class thdatass {
 
   class thsurvey * psurvey;  ///< parent survey
 
-  char * comment;  ///< Station comment.  
+  const char * comment;  ///< Station comment.  
+
+  double explored;  ///< Explored length within continuation.
   
   int flags;  ///< Station flags.
+
+  thdatass_attr_map attr;  ///< Station attributes.
   
   thdatass();  ///< Standard constructor.
   
@@ -403,6 +539,60 @@ class thdataequate {
 };
 
 
+class thdatamark {
+	
+	public:
+	
+	thobjectname station;
+
+  class thsurvey * psurvey = nullptr;  ///< parent survey
+
+  thobjectsrc srcf;  ///< Source file.
+	
+	unsigned char mark;
+	
+	thdatamark() : mark(TT_DATAMARK_TEMP) {}
+	
+};
+
+
+class thdataextend {
+	
+	public:
+	
+	thobjectname to, from, before;
+
+  class thsurvey * psurvey = nullptr;  ///< parent survey
+
+  thobjectsrc srcf;  ///< Source file.
+	
+	int extend = 0;
+	
+	double extend_ratio;
+
+	thdataextend() : extend_ratio(1.0) {}
+	
+};
+
+
+
+class thstdims {
+  
+  public:
+  
+	thobjectname station;
+
+  class thsurvey * psurvey = nullptr;  ///< parent survey
+
+  thobjectsrc srcf;  ///< Source file.
+  
+  double up, down, left, right;
+  
+  thstdims() : up(thnan), down(thnan), left(thnan), right(thnan) {}
+  
+};
+
+
 typedef std::list < thdataleg > thdataleg_list;  ///< Data leg list.
 
 typedef std::list < thdatafix > thdatafix_list;  ///< Data fix list.
@@ -411,6 +601,12 @@ typedef std::list < thdatass > thdatass_list;  ///< Data stations list.
 
 typedef std::list < thdataequate > thdataequate_list;  ///< Data equates list.
 
+typedef std::list < thdataextend > thdataextend_list;  ///< Data extend list.
+
+typedef std::list < thdatamark > thdatamark_list;  ///< Universal object name list.
+
+typedef std::list < thstdims > thstdims_list;  ///< Station dimensions.
+
 #endif
 
 
