@@ -29,10 +29,18 @@
 #include "thexception.h"
 #include "thchenc.h"
 #include "thsurvey.h"
+#include "thconfig.h"
+#include "thparse.h"
+#include "thcsdata.h"
+#include "thdata.h"
+#include "thproj.h"
+#include "thcs.h"
 
 thdataobject::thdataobject()
 {
   this->db = NULL;
+  this->cs = TTCS_LOCAL;
+  this->id = 0;
   
   this->name = "";
   this->title = "";
@@ -103,7 +111,7 @@ int thdataobject::get_cmd_nargs()
 }
 
 
-char * thdataobject::get_cmd_end()
+const char * thdataobject::get_cmd_end()
 {
   return NULL;
 }
@@ -119,19 +127,20 @@ bool thdataobject::get_cmd_ends_match(char * cmd) {
 }
 
 
-char * thdataobject::get_cmd_name()
+const char * thdataobject::get_cmd_name()
 {
   return NULL;
 }
 
 
-thcmd_option_desc thdataobject::get_cmd_option_desc(char * opts)
+thcmd_option_desc thdataobject::get_cmd_option_desc(const char * opts)
 {
   int id = thmatch_token(opts, thtt_dataobject_opt);
   switch (id) {
     case TT_DATAOBJECT_AUTHOR:
     case TT_DATAOBJECT_COPYRIGHT:
     case TT_DATAOBJECT_STNS:
+    case TT_DATAOBJECT_ATTR:
       return thcmd_option_desc(id, 2);
     default:
       return thcmd_option_desc(id);
@@ -147,7 +156,38 @@ void thdataobject::set(thcmd_option_desc cod, char ** args, int argenc, unsigned
       if (th_is_extkeyword(*args))
         this->name = this->db->strstore(*args);
       else 
-        ththrow(("invalid keyword -- %s", *args));
+        ththrow("invalid keyword -- {}", *args);
+      break;
+
+    case TT_DATAOBJECT_CS:
+      switch (this->get_class_id()) {
+        case TT_DATA_CMD:
+        case TT_LAYOUT_CMD:
+        case TT_SURFACE_CMD:
+        case TT_IMPORT_CMD:
+        case TT_SCRAP_CMD:
+          break;
+        default:
+          ththrow("coordinate system specification not valid for {}", this->get_cmd_name());
+          break;  
+      }
+      {
+        int id = thcs_parse(*args);
+        if (id == TTCS_UNKNOWN) {
+          ththrow("unknown coordinate system -- {}", *args);
+        }
+        if (this->get_class_id() != TT_DATA_CMD) {
+          this->cs = id;
+          this->cs_source = this->db->csrc;
+        } else {
+          thdata * cobj;
+          cobj = (thdata *) this;
+          if (cobj->cgroup->dl_declination_north_grid && (id == TTCS_LOCAL))
+            ththrow("grid-angle has been defined -- local CS not allowed");
+          cobj->cgroup->cs = id;
+          cobj->cgroup->cs_source = this->db->csrc;
+        }
+      }
       break;
     
     case TT_DATAOBJECT_STNS:
@@ -156,25 +196,17 @@ void thdataobject::set(thcmd_option_desc cod, char ** args, int argenc, unsigned
         case TT_SCRAP_CMD:
           break;
         default:
-          ththrow(("station-names specification not valid for %s", this->get_cmd_name()))
+          ththrow("station-names specification not valid for {}", this->get_cmd_name());
           break;  
       }
       if (strlen(args[0]) == 0)
         this->stnpref = NULL;
-      else {
-        if (th_is_extkeyword(args[0]))
-          this->stnpref = this->db->strstore(args[0]);
-        else 
-          ththrow(("invalid keyword -- %s", args[0]));
-      }
+      else
+        this->stnpref = this->db->strstore(args[0]);
       if (strlen(args[1]) == 0)
         this->stnsuff = NULL;
-      else {
-        if (th_is_extkeyword(args[1]))
-          this->stnsuff = this->db->strstore(args[1]);
-        else 
-          ththrow(("invalid keyword -- %s", args[1]));
-      }
+      else
+        this->stnsuff = this->db->strstore(args[1]);
       break;
 
     case TT_DATAOBJECT_TITLE:
@@ -186,12 +218,12 @@ void thdataobject::set(thcmd_option_desc cod, char ** args, int argenc, unsigned
         case TT_GRADE_CMD:
         case TT_LAYOUT_CMD:
           if (cod.nargs > 1)
-            ththrow(("multiple option arguments -- title"))
+            ththrow("multiple option arguments -- title");
           thencode(&(this->db->buff_enc), *args, argenc);
           this->title = this->db->strstore(this->db->buff_enc.get_buffer());
           break;
         default:
-          ththrow(("title specification not allowed for this object"));
+          ththrow("title specification not allowed for this object");
           break;
       }
       break;
@@ -201,7 +233,7 @@ void thdataobject::set(thcmd_option_desc cod, char ** args, int argenc, unsigned
         case TT_DATA_CMD:
         case TT_SCRAP_CMD:
           if (cod.nargs > 2)
-            ththrow(("too many option arguments -- author"))
+            ththrow("too many option arguments -- author");
           this->dotmp_date.parse(args[0]);
           thencode(&(this->db->buff_enc), args[1], argenc);
           this->dotmp_person.parse(this->db, this->db->buff_enc.get_buffer());
@@ -210,7 +242,7 @@ void thdataobject::set(thcmd_option_desc cod, char ** args, int argenc, unsigned
           this->author_map[this->dotmp_author].join(this->dotmp_date);
           break;
         default:
-          ththrow(("author specification not allowed for this object"));
+          ththrow("author specification not allowed for this object");
           break;
       }
       break;
@@ -220,7 +252,7 @@ void thdataobject::set(thcmd_option_desc cod, char ** args, int argenc, unsigned
         case TT_DATA_CMD:
         case TT_SCRAP_CMD:
           if (cod.nargs > 2)
-            ththrow(("too many option arguments -- copyright"))
+            ththrow("too many option arguments -- copyright");
           this->dotmp_date.parse(args[0]);
           thencode(&(this->db->buff_enc), args[1], argenc);
           this->dotmp_copyright = 
@@ -230,13 +262,25 @@ void thdataobject::set(thcmd_option_desc cod, char ** args, int argenc, unsigned
           this->copyright_map[this->dotmp_copyright].join(this->dotmp_date);
           break;
         default:
-          ththrow(("copyright specification not allowed for this object"));
+          ththrow("copyright specification not allowed for this object");
           break;
       }
       break;
+
+    case TT_DATAOBJECT_ATTR:
+      switch (this->get_class_id()) {
+        case TT_LAYOUT_CMD:
+          ththrow("attribute specification not valid for layout");
+          break;
+      }
+      if (cod.nargs != 2)
+        ththrow("invalid attribute specification -- should be <name> <value>");
+      thencode(&(this->db->buff_enc), args[1], argenc);
+      this->parse_attribute(args[0], this->db->buff_enc.get_buffer());
+      break;
         
     default:
-      ththrow(("unknown option -- %s", args[0]));
+      ththrow("unknown option -- {}", args[0]);
 
   }
 }
@@ -248,43 +292,35 @@ int thdataobject::get_id()
 }
 
 
-void thdataobject::self_delete()
-{
-  delete this;
-}
-
-
 int thdataobject::get_context()
 {
   return THCTX_SURVEY;
 }
 
 
-char * thdataobject::get_name()
+const char * thdataobject::get_name()
 {
   return this->name;
 }
 
 
-char * thdataobject::get_title()
+const char * thdataobject::get_title()
 {
   return this->title;
 }
 
 
-void thdataobject::throw_source()
+std::string thdataobject::throw_source() const
 {
-  if (strlen(thexc.get_buffer()) > 0)
-    thexc.insspf(" -- ");
-  thexc.insspf("%s [%d]", this->source.name, this->source.line);
+  return fmt::format("{} [{}]", source.name, source.line);
 }
 
 void thdataobject::self_print(FILE * outf)
 {
   if (strlen(this->name) > 0)
-    fprintf(outf,"%s (%ld:0x%lx) -- %s\n", this->get_class_name(), this->id, (unsigned long) this, this->name);
+    fprintf(outf,"%s (%ld:0x%lx) -- %s\n", this->get_class_name(), this->id, (intptr_t) this, this->name);
   else
-    fprintf(outf,"%s (%ld:0x%lx)\n", this->get_class_name(), this->id, (unsigned long) this);  
+    fprintf(outf,"%s (%ld:0x%lx)\n", this->get_class_name(), this->id, (intptr_t) this);  
 
   this->self_print_properties(outf);
 
@@ -368,3 +404,140 @@ void thdataobject::start_insert()
 }
 
 
+void thdataobject::read_cs(char * src_x, char * src_y, double & dst_x, double & dst_y, bool adj_bbox)
+{
+	  // 1. Check compatibility with output CS.
+	  if (thcfg.outcs_def.is_valid()) {
+	    if (((this->cs == TTCS_LOCAL) && (thcfg.outcs != TTCS_LOCAL)) ||
+	      ((this->cs != TTCS_LOCAL) && (thcfg.outcs == TTCS_LOCAL)))
+	      ththrow("mixing local and global coordinate systems not allowed -- conflict with cs specification at {} [{}]", thcfg.outcs_def.name, thcfg.outcs_def.line);
+	  };
+
+	  // 1. Conversion to numbers.
+	  int sv;
+	  double tx(0.0), ty(0.0), tz(0.0);
+	  if ((this->cs != TTCS_LOCAL) && thcs_get_data(this->cs)->dms) {
+	    thparse_double_dms(sv, tx, src_x);
+	    tx /= 180.0 / THPI;
+	  } else {
+	    thparse_double(sv, tx, src_x);
+	  }
+	  if (sv != TT_SV_NUMBER)
+	    ththrow("invalid X coordinate -- {}", src_x);
+
+	  if ((this->cs != TTCS_LOCAL) && thcs_get_data(this->cs)->dms) {
+	    thparse_double_dms(sv, ty, src_y);
+	    ty /= 180.0 / THPI;
+	  } else {
+	    thparse_double(sv, ty, src_y);
+	  }
+	  if (sv != TT_SV_NUMBER)
+	    ththrow("invalid Y coordinate -- {}", src_y);
+
+	  if ((this->cs != TTCS_LOCAL) && thcs_get_data(this->cs)->swap) {
+	    tz = tx;
+	    tx = ty;
+	    ty = tz;
+	    tz = 0.0;
+	  }
+
+	  if ((this->cs != TTCS_LOCAL) && thcs_get_data(this->cs)->dms) {
+	    if ((tx < - THPI) || (tx > THPI))
+	      ththrow("longitude out of range -- {}", thcs_get_data(this->cs)->swap ? src_y : src_x);
+
+	    if ((ty < (- THPI / 2)) || (ty > (THPI / 2)))
+	      ththrow("latitude out of range -- {}", thcs_get_data(this->cs)->swap ? src_x : src_y);
+	  }
+
+	  dst_x = tx;
+	  dst_y = ty;
+
+	  if (adj_bbox && (this->cs != TTCS_LOCAL)) {
+		double dumx, dumy, dumz;
+		thcs2cs(thcs_get_params(this->cs), thcs_get_params(TTCS_LAT_LONG), tx, ty, tz, dumx, dumy, dumz);
+	    if (!thcfg.ibbx_def) {
+	      thcfg.ibbx[0] = dumx;
+	      thcfg.ibbx[1] = dumx;
+	      thcfg.ibbx[2] = dumy;
+	      thcfg.ibbx[3] = dumy;
+	      thcfg.ibbx_def = true;
+	    } else {
+	      if (dumx < thcfg.ibbx[0]) thcfg.ibbx[0] = dumx;
+	      if (dumx > thcfg.ibbx[1]) thcfg.ibbx[1] = dumx;
+	      if (dumy < thcfg.ibbx[2]) thcfg.ibbx[2] = dumy;
+	      if (dumy > thcfg.ibbx[3]) thcfg.ibbx[3] = dumy;
+	    }
+	  }
+
+}
+
+
+
+void thdataobject::convert_cs(int src_cs, double src_x, double src_y, double & dst_x, double & dst_y)
+{
+  bool initcs(false);
+  double tx(0.0), ty(0.0), tz(0.0), dst_z(0.0);
+  tx = src_x;
+  ty = src_y;
+
+  if (!thcfg.outcs_def.is_valid()) {
+    if ((src_cs != TTCS_LOCAL) && (!thcs_get_data(src_cs)->output)) {
+      // TODO: get NS
+      double dumx, dumy, dumz;
+      int south = 0;
+      thcs2cs(thcs_get_params(src_cs), thcs_get_params(TTCS_LAT_LONG), tx, ty, tz, dumx, dumy, dumz);
+      if (dumy < 0.0)
+        south = 1;
+      thcfg.outcs = TTCS_UTM1N + 2 * (thcs2zone(thcs_get_params(src_cs), tx, ty, tz) - 1) + south;
+    } else {
+      thcfg.outcs = src_cs;
+    }
+    if (this->cs_source.is_valid())
+      thcfg.outcs_def = this->cs_source;
+    else
+      thcfg.outcs_def = this->source;
+    initcs = true;
+  }
+
+  if (src_cs == TTCS_LOCAL) {
+    dst_x = tx;
+    dst_y = ty;
+    dst_z = tz;
+  } else {
+    thcs2cs(thcs_get_params(src_cs), thcs_get_params(thcfg.outcs), tx, ty, tz, dst_x, dst_y, dst_z);
+  }
+
+  if (thcfg.outcs != TTCS_LOCAL) {
+    if (initcs) {
+      thcfg.outcs_sumx = dst_x;
+      thcfg.outcs_sumy = dst_y;
+      thcfg.outcs_sumz = dst_z;
+      thcfg.outcs_sumn = 1.0;
+    } else {
+      thcfg.outcs_sumx += dst_x;
+      thcfg.outcs_sumy += dst_y;
+      thcfg.outcs_sumz += dst_z;
+      thcfg.outcs_sumn += 1.0;
+    }
+  }
+  
+}
+
+
+void thdataobject::parse_attribute(char * name, char * value) {
+
+  // check name
+  if ((name == NULL) || (strlen(name) == 0))
+    ththrow("empty attribute name not allowed");
+  if (name[0] == '_')
+    ththrow("attribute name starting with '_' not allowed");
+  if (!th_is_attr_name(name))
+    ththrow("invalid characters in attribute name -- {}", name);
+
+  this->db->attr.insert_attribute(name, value, long(this->id));
+}
+
+
+void thdataobject::convert_all_cs() {
+}
+
